from abc import ABC, abstractmethod
from enum import IntEnum
from typing import List, NamedTuple

from geoalchemy2.functions import ST_SRID
from sqlalchemy import and_, false, func, types
from sqlalchemy.orm.session import Session
from threedi_schema.domain import custom_types


class CheckLevel(IntEnum):
    ERROR = 40
    FUTURE_ERROR = 39
    WARNING = 30
    INFO = 20

    @classmethod
    def get(cls, value):
        """Get a CheckLevel from a CheckLevel, str or int."""
        if isinstance(value, cls):
            return value
        elif isinstance(value, str):
            return cls[value.upper()]
        else:
            return cls(value)


class BaseCheck(ABC):
    """Base class for all checks.

    A Check defines a constraint on a specific column and its table.
    One can validate if the constrain holds using the method `get_invalid()`.
    This method will return a list of rows (as named_tuples) which are invalid.
    """

    def __init__(
        self,
        column,
        filters=None,
        level=CheckLevel.ERROR,
        error_code=0,
        is_beta_check=False,
    ):
        self.column = column
        self.table = column.table
        self.filters = filters
        self.error_code = int(error_code)
        self.level = CheckLevel.get(level)
        self.is_beta_check = is_beta_check

    @abstractmethod
    def get_invalid(self, session: Session) -> List[NamedTuple]:
        """Return a list of rows (named_tuples) which are invalid.

        What is invalid is defined in the check. Returns an empty list if no
        invalid rows are present.

        :param session: sqlalchemy.orm.session.Session
        :return: list of named_tuples or empty list if there are no invalid
            rows
        """
        pass

    def get_valid(self, session: Session) -> List[NamedTuple]:
        """Return a list of rows (named_tuples) which are valid.

        :param session: sqlalchemy.orm.session.Session
        :return: list of named_tuples or empty list if there are no valid rows
        """
        all_rows = self.to_check(session)
        invalid_row_ids = set([row.id for row in self.get_invalid(session)])
        valid = []
        for row in all_rows:
            if row.id not in invalid_row_ids:
                valid.append(row)
        return valid

    def to_check(self, session):
        """Return a Query object filtering on the rows this check is applied.

        :param session: sqlalchemy.orm.session.Session
        :return: sqlalchemy.Query
        """
        query = session.query(self.table)
        if self.filters is not None:
            query = query.filter(self.filters)
        return query

    @property
    def column_name(self) -> str:
        return f"{self.table.name}.{self.column.name}"

    def description(self) -> str:
        """Return a string explaining why rows are invalid according to this
        check.

        :return: string
        """
        return "Invalid value in column '%s'" % self.column_name

    def __repr__(self) -> str:
        return "<%s: %s>" % (self.__class__.__name__, self.column_name)


class QueryCheck(BaseCheck):
    """Specify a sqlalchemy.orm.Query object to return invalid instances"""

    def __init__(
        self,
        column,
        invalid,
        message,
        filters=None,
        level=CheckLevel.ERROR,
        error_code=0,
        is_beta_check=False,
    ):
        super().__init__(
            column, level=level, error_code=error_code, is_beta_check=is_beta_check
        )
        self.invalid = invalid
        self.message = message
        self.filters = filters

    def get_invalid(self, session):
        query = self.invalid.with_session(session)
        if self.filters is not None:
            query = query.filter(self.filters)
        return query.all()

    def description(self):
        return self.message


class ForeignKeyCheck(BaseCheck):
    """Check all values in `column` are in `reference_column`.

    Null values are ignored."""

    def __init__(self, reference_column, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.reference_column = reference_column

    def get_invalid(self, session):
        q_invalid = self.to_check(session)
        invalid_foreign_keys_query = q_invalid.filter(
            self.column.notin_(session.query(self.reference_column)),
            self.column != None,
        )
        return invalid_foreign_keys_query.all()

    def description(self):
        return "%s refers to a non-existing %s" % (
            self.column_name,
            self.reference_column.table,
        )


class UniqueCheck(BaseCheck):
    """Check all values in `column` are unique

    Null values are ignored."""

    def __init__(self, columns, **kwargs):
        if not isinstance(columns, (list, tuple)):
            columns = (columns,)
        self.columns = columns
        super().__init__(column=columns[0], **kwargs)

    def get_invalid(self, session):
        duplicate_values = (
            session.query(*self.columns)
            .group_by(*self.columns)
            .having(func.count(self.column) > 1)
        ).subquery()
        join_clause = and_(
            *[getattr(duplicate_values.c, c.name) == c for c in self.columns]
        )
        q_invalid = self.to_check(session).join(duplicate_values, join_clause)
        return q_invalid.all()

    def description(self):
        if len(self.columns) > 1:
            return f"columns {sorted({c.name for c in self.columns})} in table {self.table.name} should be unique together"
        else:
            return f"{self.column_name} should be unique"


class AllEqualCheck(BaseCheck):
    """Check all values in `column` are the same, including NULL values."""

    def get_invalid(self, session):
        val = session.query(self.column).limit(1).scalar()
        if val is None:
            clause = self.column != None
        else:
            clause = (self.column != val) | (self.column == None)
        return self.to_check(session).filter(clause).all()

    def description(self):
        return f"{self.column_name} is different and is ignored if it is not in the first record"


class NotNullCheck(BaseCheck):
    """ "Check all values in `column` that are not null"""

    def get_invalid(self, session):
        q_invalid = self.to_check(session)
        not_null_query = q_invalid.filter(self.column == None)
        return not_null_query.all()

    def description(self):
        return f"{self.column_name} cannot be null"


class TypeCheck(BaseCheck):
    """Check all values in `column` that are of the column defined type.

    Null values are ignored."""

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.expected_types = _sqlalchemy_to_sqlite_types(self.column.type)

    def get_invalid(self, session):
        if ("sqlite" not in session.bind.dialect.dialect_description) and (
            "geopackage" not in session.bind.dialect.dialect_description
        ):
            return []
        q_invalid = self.to_check(session)
        invalid_type_query = q_invalid.filter(
            func.typeof(self.column).notin_(self.expected_types),
            func.typeof(self.column) != "null",
        )
        return invalid_type_query.all()

    def description(self):
        return f"{self.column_name} is not of type {self.expected_types}"


def _sqlalchemy_to_sqlite_types(column_type):
    """Convert the sqlalchemy column type to allowed sqlite data types

    Returns the value similar as the sqlite 'typeof' function.
    Raises TypeError if the column type is unknown.
    See https://www.sqlite.org/datatype3.html

    :param column_type: sqlalchemy.column
    :return: (str)
    """
    if isinstance(column_type, types.TypeDecorator):
        column_type = column_type.impl

    if isinstance(column_type, types.String):
        return ["text"]
    elif isinstance(column_type, (types.Float, types.Numeric)):
        return ["integer", "numeric", "real"]
    elif isinstance(column_type, types.Integer):
        return ["integer"]
    elif isinstance(column_type, types.Boolean):
        return ["integer"]
    elif isinstance(column_type, types.Date):
        return ["text"]
    elif isinstance(column_type, custom_types.Geometry):
        return ["blob"]
    elif isinstance(column_type, types.TIMESTAMP):
        return ["text"]
    else:
        raise TypeError("Unknown column type: %s" % column_type)


class GeometryCheck(BaseCheck):
    """Check all values in `column` are a valid geometry.

    Null values are ignored."""

    def get_invalid(self, session):
        q_invalid = self.to_check(session)
        invalid_geometries = q_invalid.filter(
            func.ST_IsValid(self.column) != True, self.column != None
        )
        return invalid_geometries.all()

    def description(self):
        return f"{self.column_name} is an invalid geometry"


class GeometryTypeCheck(BaseCheck):
    """Check all values in `column` are of geometry type in defined in
    `column`.

    Null values are ignored"""

    def get_invalid(self, session):
        expected_geometry_type = _get_geometry_type(
            self.column, dialect=session.bind.dialect.name
        )
        q_invalid = self.to_check(session)
        if expected_geometry_type is None:
            # skip in case of generic GEOMETRY column
            return q_invalid.filter(false())
        invalid_geometry_types_q = q_invalid.filter(
            func.ST_GeometryType(self.column) != expected_geometry_type,
            self.column != None,
        )
        return invalid_geometry_types_q.all()

    def description(self):
        return "%s has invalid geometry type, expected %s" % (
            self.column_name,
            self.column.type.geometry_type,
        )


def _get_geometry_type(column, dialect):
    if column.type.geometry_type == "GEOMETRY":
        return  # should skip the check
    if dialect in ["sqlite", "geopackage"]:
        return column.type.geometry_type
    elif dialect == "postgresql":
        geom_type = column.type.geometry_type.capitalize()
        return "ST_%s" % geom_type
    else:
        raise TypeError("Unexpected dialect %s" % dialect)


class EnumCheck(BaseCheck):
    """Check all values in `column` are within the defined Enum values of
    `column`.

    Unexpected values are values not defined by its enum_class.

    Null values are ignored"""

    def get_invalid(self, session):
        q_invalid = self.to_check(session)
        invalid_values_q = q_invalid.filter(
            self.column.notin_(list(self.column.type.enum_class))
        )
        return invalid_values_q.all()

    def description(self):
        allowed = sorted({x.value for x in self.column.type.enum_class})
        return f"{self.column_name} is not one of {allowed}"


class RangeCheck(BaseCheck):
    """Check to if all values are within range [min_value, max_value]

    Use left_inclusive and right_inclusive to specify whether the min/max values
    themselves should be considered valid. By default they are both considered
    valid.
    """

    def __init__(
        self,
        min_value=None,
        max_value=None,
        left_inclusive=True,
        right_inclusive=True,
        message=None,
        *args,
        **kwargs,
    ):
        if min_value is None and max_value is None:
            raise ValueError("Please supply at least one of {min_value, max_value}.")
        self.min_value = min_value
        self.max_value = max_value
        self.left_inclusive = left_inclusive
        self.right_inclusive = right_inclusive
        self.message = message
        super().__init__(*args, **kwargs)

    def get_invalid(self, session):
        conditions = []
        if self.min_value is not None:
            if self.left_inclusive:
                conditions.append(self.column >= self.min_value)
            else:
                conditions.append(self.column > self.min_value)
        if self.max_value is not None:
            if self.right_inclusive:
                conditions.append(self.column <= self.max_value)
            else:
                conditions.append(self.column < self.max_value)
        return self.to_check(session).filter(~and_(*conditions)).all()

    def description(self):
        if self.message:
            return self.message
        parts = []
        if self.min_value is not None:
            parts.append(f"{'<' if self.left_inclusive else '<='}{self.min_value}")
        if self.max_value is not None:
            parts.append(f"{'>' if self.right_inclusive else '>='}{self.max_value}")
        return f"{self.column_name} is {' and/or '.join(parts)}"


class ListOfIntsCheck(BaseCheck):
    def get_invalid(self, session):
        invalids = []
        for record in self.to_check(session).filter(
            (self.column != None) & (self.column != "")
        ):
            # check if casting to int works
            try:
                [int(x) for x in getattr(record, self.column.name).split(",")]
            except ValueError:
                invalids.append(record)
        return invalids

    def description(self) -> str:
        return (
            f"{self.table.name}.{self.column} is not a comma seperated list of integers"
        )


class EPSGGeomCheck(BaseCheck):
    def __init__(
        self,
        *args,
        **kwargs,
    ):
        super().__init__(*args, **kwargs)
        self.epsg_ref_name = ""

    def get_invalid(self, session: Session) -> List[NamedTuple]:
        self.epsg_ref_name = session.model_checker_context.epsg_ref_name
        if session.model_checker_context.epsg_ref_code is None:
            return []
        return (
            self.to_check(session)
            .filter(ST_SRID(self.column) != session.model_checker_context.epsg_ref_code)
            .all()
        )

    def description(self) -> str:
        return f"The epsg of {self.table.name}.{self.column_name} should match {self.epsg_ref_name}"



from inspect import isclass

import factory
from factory import Faker
from threedi_schema import constants, models

# Default geometry strings

DEFAULT_POINT = "SRID=28992;POINT (142742 473443)"
# line from DEFAULT_POINT to another point
DEFAULT_LINE = "SRID=28992;LINESTRING (142742 473443, 142747 473448)"
# polygon containing DEFAULT_POINT
DEFAULT_POLYGON = "SRID=28992;POLYGON ((142742 473443, 142743 473443, 142744 473444, 142742 473444, 142742 473443))"


def inject_session(session):
    """Inject the session into all factories"""
    for _, cls in globals().items():
        if isclass(cls) and issubclass(cls, factory.alchemy.SQLAlchemyModelFactory):
            cls._meta.sqlalchemy_session = session


class BaseFactory(factory.alchemy.SQLAlchemyModelFactory):
    class Meta:
        abstract = True

    id = factory.Sequence(lambda n: n + 1)


class TimeStepSettingsFactory(BaseFactory):
    class Meta:
        model = models.TimeStepSettings
        sqlalchemy_session = None

    time_step = 30
    min_time_step = 1
    max_time_step = 100
    output_time_step = 300
    use_time_step_stretch = False


class ModelSettingsFactory(BaseFactory):
    class Meta:
        model = models.ModelSettings
        sqlalchemy_session = None

    friction_averaging = 0
    minimum_cell_size = 20
    calculation_point_distance_1d = 15
    minimum_table_step_size = 0.05
    use_1d_flow = False
    use_2d_rain = 1
    nr_grid_levels = 4
    friction_coefficient = 0.03
    use_2d_flow = True
    friction_type = constants.FrictionType.CHEZY


class ConnectionNodeFactory(BaseFactory):
    class Meta:
        model = models.ConnectionNode
        sqlalchemy_session = None

    code = Faker("name")
    geom = DEFAULT_POINT


class ChannelFactory(BaseFactory):
    class Meta:
        model = models.Channel
        sqlalchemy_session = None

    display_name = Faker("name")
    code = Faker("name")
    exchange_type = constants.CalculationType.CONNECTED
    geom = DEFAULT_LINE


class WeirFactory(BaseFactory):
    class Meta:
        model = models.Weir
        sqlalchemy_session = None

    code = factory.Sequence(lambda n: "Code %d" % n)
    display_name = "display_name"
    crest_level = 1.0
    crest_type = constants.CrestType.BROAD_CRESTED
    friction_value = 2.0
    friction_type = constants.FrictionType.CHEZY
    sewerage = False
    connection_node_id_start = 1
    connection_node_id_end = 1
    geom = DEFAULT_LINE


class BoundaryConditions2DFactory(BaseFactory):
    class Meta:
        model = models.BoundaryConditions2D
        sqlalchemy_session = None

    type = constants.BoundaryType.WATERLEVEL.value
    timeseries = "0,-0.5"
    display_name = Faker("name")
    geom = DEFAULT_LINE


class BoundaryConditions1DFactory(BaseFactory):
    class Meta:
        model = models.BoundaryCondition1D
        sqlalchemy_session = None

    type = constants.BoundaryType.WATERLEVEL
    timeseries = "0,-0.5"
    connection_node_id = 1
    geom = DEFAULT_POINT


class PumpMapFactory(BaseFactory):
    class Meta:
        model = models.PumpMap
        sqlalchemy_session = None

    geom = DEFAULT_POINT


class PumpFactory(BaseFactory):
    class Meta:
        model = models.Pump
        sqlalchemy_session = None

    code = "code"
    display_name = "display_name"
    sewerage = False
    type_ = constants.PumpType.DELIVERY_SIDE
    start_level = 1.0
    lower_stop_level = 0.0
    capacity = 5.0
    geom = DEFAULT_POINT


class CrossSectionLocationFactory(BaseFactory):
    class Meta:
        model = models.CrossSectionLocation
        sqlalchemy_session = None

    code = "code"
    reference_level = 0.0
    friction_type = constants.FrictionType.CHEZY
    friction_value = 0.0
    geom = DEFAULT_POINT


class AggregationSettingsFactory(BaseFactory):
    class Meta:
        model = models.AggregationSettings
        sqlalchemy_session = None

    flow_variable = "waterlevel"
    aggregation_method = "avg"
    interval = 10


class NumericalSettingsFactory(BaseFactory):
    class Meta:
        model = models.NumericalSettings
        sqlalchemy_session = None

    max_degree_gauss_seidel = 1
    use_of_cg = 20
    use_nested_newton = 0
    flooding_threshold = 0.01


class Lateral1DFactory(BaseFactory):
    class Meta:
        model = models.Lateral1D
        sqlalchemy_session = None

    timeseries = "0,-0.1"
    connection_node_id = 1
    geom = DEFAULT_POINT


class Lateral2DFactory(BaseFactory):
    class Meta:
        model = models.Lateral2D
        sqlalchemy_session = None

    timeseries = "0,-0.2"
    geom = DEFAULT_POINT
    type = constants.Later2dType.SURFACE


class SurfaceParametersFactory(BaseFactory):
    class Meta:
        model = models.SurfaceParameters
        sqlalchemy_session = None

    outflow_delay = 10.0
    surface_layer_thickness = 5.0
    infiltration = True
    max_infiltration_capacity = 10.0
    min_infiltration_capacity = 5.0
    infiltration_decay_constant = 3.0
    infiltration_recovery_constant = 2.0


class DryWheatherFlowMapFactory(BaseFactory):
    class Meta:
        model = models.DryWeatherFlowMap
        sqlalchemy_session = None

    geom = DEFAULT_POINT


class DryWeatherFlowFactory(BaseFactory):
    class Meta:
        model = models.DryWeatherFlow
        sqlalchemy_session = None

    geom = DEFAULT_POLYGON


class DryWeatherFlowDistributionFactory(BaseFactory):
    class Meta:
        model = models.DryWeatherFlowDistribution
        sqlalchemy_session = None

    id = 1
    distribution = (
        "3,1.5,1,1,0.5,0.5,2.5,8,7.5,6,5.5,5,4.5,4,4,3.5,3.5,4,5.5,8,7,5.5,4.5,4"
    )


class SurfaceFactory(BaseFactory):
    class Meta:
        model = models.Surface
        sqlalchemy_session = None

    area = 0.0
    geom = DEFAULT_POLYGON

    # surface_parameters = factory.SubFactory(SurfaceParameterFactory)


class SurfaceMapFactory(BaseFactory):
    class Meta:
        model = models.SurfaceMap
        sqlalchemy_session = None

    percentage = 100.0
    geom = DEFAULT_LINE


class TagsFactory(BaseFactory):
    class Meta:
        model = models.Tags
        sqlalchemy_session = None


class TableControlFactory(BaseFactory):
    class Meta:
        model = models.TableControl
        sqlalchemy_session = None

    action_type = constants.TableControlActionTypes.set_discharge_coefficients
    action_table = "0.0,-1.0 2.0\n1.0,-1.1 2.1"
    measure_operator = constants.MeasureOperators.greater_than
    target_type = constants.StructureControlTypes.channel
    target_id = 10
    geom = DEFAULT_POINT


class MemoryControlFactory(BaseFactory):
    class Meta:
        model = models.MemoryControl
        sqlalchemy_session = None

    action_type = constants.TableControlActionTypes.set_discharge_coefficients
    action_value_1 = 0.0
    action_value_2 = -1.0
    target_type = constants.StructureControlTypes.channel
    target_id = 10
    is_inverse = False
    is_active = True
    upper_threshold = 1.0
    lower_threshold = -1.0
    geom = DEFAULT_POINT


class MeasureMapFactory(BaseFactory):
    class Meta:
        model = models.MeasureMap
        sqlalchemy_session = None

    control_type = constants.MeasureVariables.waterlevel
    geom = DEFAULT_LINE


class MeasureLocationFactory(BaseFactory):
    class Meta:
        model = models.MeasureLocation
        sqlalchemy_session = None

    geom = DEFAULT_POINT


class CulvertFactory(BaseFactory):
    class Meta:
        model = models.Culvert
        sqlalchemy_session = None

    code = "code"
    display_name = Faker("name")
    exchange_type = constants.CalculationTypeCulvert.ISOLATED_NODE
    geom = DEFAULT_LINE
    friction_value = 0.03
    friction_type = 2
    invert_level_start = 0.1
    invert_level_end = 1.1
    discharge_coefficient_negative = 1.0
    discharge_coefficient_positive = 1.0


class PotentialBreachFactory(BaseFactory):
    class Meta:
        model = models.PotentialBreach
        sqlalchemy_session = None

    display_name = Faker("name")
    code = "code"
    geom = DEFAULT_LINE


class VegetationDragFactory(BaseFactory):
    class Meta:
        model = models.VegetationDrag2D
        sqlalchemy_session = None

    vegetation_height = 1.0
    vegetation_height_file = "vegetation_height_file.txt"

    vegetation_stem_count = 50000
    vegetation_stem_count_file = "vegetation_stem_count_file.txt"

    vegetation_stem_diameter = 0.5
    vegetation_stem_diameter_file = "vegetation_stem_diameter_file.txt"

    vegetation_drag_coefficient = 0.4
    vegetation_drag_coefficient_file = "vegetation_drag_coefficient_file.txt"


class SimulationTemplateSettingsFactory(BaseFactory):
    class Meta:
        model = models.SimulationTemplateSettings
        sqlalchemy_session = None

    name = "Foo"
    use_0d_inflow = constants.InflowType.NO_INFLOW



from typing import Dict, Iterator, NamedTuple, Optional, Tuple

from threedi_schema import models, ThreediDatabase

from .checks.base import BaseCheck, CheckLevel
from .checks.raster import LocalContext, ServerContext
from .config import Config

__all__ = ["ThreediModelChecker"]


def get_epsg_data_from_raster(session) -> Tuple[int, str]:
    """
    Retrieve epsg code for schematisation loaded in session. This is done by
    iterating over all geometries in the declared models and all raster files, and
    stopping at the first geometry or raster file with data.

    Returns the epsg code and the name (table.column) of the source.
    """
    context = session.model_checker_context
    raster_interface = context.raster_interface if context is not None else None
    epsg_code = None
    epsg_source = ""
    raster = models.ModelSettings.dem_file
    raster_files = session.query(raster).filter(raster != None, raster != "").first()
    if raster_files is not None and raster_files is not None:
        if isinstance(context, ServerContext):
            if isinstance(context.available_rasters, dict):
                abs_path = context.available_rasters.get(raster.name)
        else:
            abs_path = context.base_path.joinpath("rasters", raster_files[0])
        with raster_interface(abs_path) as ro:
            if ro.epsg_code is not None:
                epsg_code = ro.epsg_code
                epsg_source = "model_settings.dem_file"
    return epsg_code, epsg_source


class ThreediModelChecker:
    def __init__(
        self,
        threedi_db: ThreediDatabase,
        context: Optional[Dict] = None,
        allow_beta_features=False,
    ):
        """Initialize the model checker.

        Optionally, supply the context of the model check:

        - "context_type": "local" or "server", default "local"
        - "raster_interface": a threedi_modelchecker.interfaces.RasterInterface subclass
        - "base_path": (only local) path where to look for rasters (defaults to the db's directory)
        - "available_rasters": (only server) a dict of raster_option -> raster url
        """
        self.db = threedi_db
        self.schema = self.db.schema
        self.schema.validate_schema()
        self.config = Config(
            models=self.models, allow_beta_features=allow_beta_features
        )
        context = {} if context is None else context.copy()
        context_type = context.pop("context_type", "local")
        session = self.db.get_session()
        if self.db.schema.epsg_code is not None:
            context["epsg_ref_code"] = self.db.schema.epsg_code
            context["epsg_ref_name"] = self.db.schema.epsg_source
        else:
            epsg_ref_code, epsg_ref_name = get_epsg_data_from_raster(session)
            context["epsg_ref_code"] = epsg_ref_code
            context["epsg_ref_name"] = epsg_ref_name

        if context_type == "local":
            context.setdefault("base_path", self.db.base_path)
            self.context = LocalContext(**context)
        elif context_type == "server":
            self.context = ServerContext(**context)
        else:
            raise ValueError(f"Unknown context_type '{context_type}'")

        session.model_checker_context = self.context

    @property
    def models(self):
        """Returns a list of declared models"""
        return self.schema.declared_models

    def errors(
        self, level=CheckLevel.ERROR, ignore_checks=None
    ) -> Iterator[Tuple[BaseCheck, NamedTuple]]:
        """Iterates and applies checks, returning any failing rows.

        By default, checks of WARNING and INFO level are ignored.

        :return: Tuple of the applied check and the failing row.
        """

        session = self.db.get_session()
        session.model_checker_context = self.context
        for check in self.checks(level=level, ignore_checks=ignore_checks):
            model_errors = check.get_invalid(session)
            for error_row in model_errors:
                yield check, error_row

    def checks(self, level=CheckLevel.ERROR, ignore_checks=None) -> Iterator[BaseCheck]:
        """Iterates over all configured checks

        :return: implementations of BaseChecks
        """
        for check in self.config.iter_checks(level=level, ignore_checks=ignore_checks):
            yield check

    def check_table(self, table):
        pass

    def check_column(self, column):
        pass



import pytest
from threedi_schema import models

from threedi_modelchecker.checks.base import CheckLevel
from threedi_modelchecker.checks.factories import (
    ForeignKeyCheckSetting,
    generate_enum_checks,
    generate_epsg_geom_checks,
    generate_epsg_raster_checks,
    generate_foreign_key_checks,
    generate_geometry_checks,
    generate_not_null_checks,
    generate_unique_checks,
)


def test_gen_foreign_key_checks_no_filter():
    settings = [
        ForeignKeyCheckSetting(
            models.Surface.surface_parameters_id, models.SurfaceParameters.id
        ),
        ForeignKeyCheckSetting(
            models.MemoryControl.target_id,
            models.Channel.id,
            models.MemoryControl.target_type == "channel",
        ),
    ]
    foreign_key_checks = generate_foreign_key_checks(
        models.Surface.__table__, fk_settings=settings
    )
    assert len(foreign_key_checks) == 1
    fk_check = foreign_key_checks[0]
    assert models.Surface.surface_parameters_id == fk_check.column
    assert models.SurfaceParameters.id == fk_check.reference_column


def test_gen_foreign_key_checks_filter():
    settings = [
        ForeignKeyCheckSetting(
            models.Surface.surface_parameters_id, models.SurfaceParameters.id
        ),
        ForeignKeyCheckSetting(
            models.MemoryControl.target_id,
            models.Channel.id,
            models.MemoryControl.target_type == "channel",
        ),
    ]
    foreign_key_checks = generate_foreign_key_checks(
        models.MemoryControl.__table__, fk_settings=settings
    )
    assert len(foreign_key_checks) == 1
    fk_check = foreign_key_checks[0]
    assert settings[1].filter == fk_check.filters


def test_gen_not_unique_checks():
    not_unique_checks = generate_unique_checks(models.Channel.__table__)
    assert len(not_unique_checks) == 1
    assert models.Channel.id == not_unique_checks[0].column


@pytest.mark.parametrize(
    "extra_columns", [None, [], [models.Channel.connection_node_id_end]]
)
def test_gen_not_null_checks(extra_columns):
    not_null_checks = generate_not_null_checks(
        models.Channel.__table__, extra_not_null_columns=extra_columns
    )
    if extra_columns is None:
        extra_columns = []
    assert len(not_null_checks) == 2 + len(extra_columns)
    not_null_check_columns = [check.column for check in not_null_checks]
    assert models.Channel.id in not_null_check_columns
    assert models.Channel.geom in not_null_check_columns
    for col in extra_columns:
        assert col in not_null_check_columns
    assert models.Channel.code not in not_null_check_columns


def test_gen_geometry_check():
    geometry_checks = generate_geometry_checks(models.ConnectionNode.__table__)

    assert len(geometry_checks) == 1
    geometry_check_columns = [check.column for check in geometry_checks]
    assert models.ConnectionNode.geom in geometry_check_columns


def test_gen_enum_checks():
    enum_checks = generate_enum_checks(models.BoundaryConditions2D.__table__)

    assert len(enum_checks) == 1
    assert enum_checks[0].column == models.BoundaryConditions2D.type


def test_gen_enum_checks_varcharenum():
    enum_checks = generate_enum_checks(models.AggregationSettings.__table__)

    assert len(enum_checks) == 2
    enum_check_columns = [check.column for check in enum_checks]
    assert models.AggregationSettings.aggregation_method in enum_check_columns
    assert models.AggregationSettings.flow_variable in enum_check_columns


@pytest.mark.parametrize(
    "name", ["*.aggregation_method", "aggregation_settings.aggregation_method"]
)
def test_gen_enum_checks_custom_mapping(name):
    enum_checks = generate_enum_checks(
        models.AggregationSettings.__table__,
        custom_level_map={name: "WARNING"},
    )

    assert len(enum_checks) == 2
    checks = {check.column.name: check for check in enum_checks}
    assert checks["aggregation_method"].level == CheckLevel.WARNING
    assert checks["flow_variable"].level == CheckLevel.ERROR


@pytest.mark.parametrize(
    "model, nof_checks_expected",
    [
        (models.ModelSettings, 0),
        (models.ConnectionNode, 1),
    ],
)
def test_gen_epsg_geom_checks(model, nof_checks_expected):
    assert len(generate_epsg_geom_checks(model.__table__)) == nof_checks_expected


@pytest.mark.parametrize(
    "model, nof_checks_expected",
    [
        (models.ConnectionNode, 0),
        (models.GroundWater, 1),
        (models.ModelSettings, 2),
    ],
)
def test_gen_epsg_raster_checks(model, nof_checks_expected):
    raster_columns = [
        models.ModelSettings.dem_file,
        models.ModelSettings.friction_coefficient_file,
        models.GroundWater.equilibrium_infiltration_rate_file,
    ]
    assert (
        len(generate_epsg_raster_checks(model.__table__, raster_columns))
        == nof_checks_expected
    )
